/* GSI - general sound interface   [-*- C++ -*-]
 *
 * dsp.c - shared sound server, dsp mode setting code 
 *   - sept 1996/may 1997/dec 1997/march-may 1998
 *
 * Copyright (C) 1996,1997,1998 W.H.Scholten
 *
 * Lower level API dependencies:
 *   - DSP mode setting: OSS but should not be too difficult to adapt
 *     to other systems.
 *   - OSS include locations are Linux/FreeBSD specific.
 *     Locations for other OS's?
 *   - sequencer/synth (MIDI's) playing: very OSS specific.
 */

/* This software is licensed under the CGPL version 0.4 or later
 * (see the file LICENSE.CGPL for details)
 */




/*__________________INCLUDES___________________*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <errno.h>

#include <math.h>

#include <gsi/config.h>
#ifdef USE_OSS
#ifdef __linux__
#  include <sys/soundcard.h>
#endif
#ifdef __FreeBSD__
#  include <machine/soundcard.h>
#  include <sys/errno.h>
#endif
#endif
#ifdef USE_NEW_BSD
#  error code for new bsd driver not implemented
#endif
#ifdef USE_DEV_AUDIO
#  warning some includes for /dev/audio may be missing
#endif

#include <gsi/mytypes.h>
#include <gsi/debug.h>
#define _SNDSERVER_
#include <gsi/gsi_commands.h>
#include <gsi/gsi_server_version.h>
#include "midi.h"

/*__________________DEFINITIES___________________*/

#define min(x,y) ((x)>(y) ? (y): (x))
#define max(x,y) ((x)<(y) ? (y): (x))


/*__________________EXTERN___________________*/
extern int  dsp;

extern int  dsp_speed;
extern int  dsp_type;
extern int  dsp_log2_frag;
extern int  dsp_fragsize; /* samples=fragsize/channels/bytes */
extern int  dsp_channels;
extern int  dsp_bps; /* bytes per sample, 1 or 2 (determined from dsp_type) */

extern int  dsp_samples; // number of samples that fit in the fragsize

extern char *server_name;

/*__________________VARIABELEN___________________*/

/*__________________FUNCTIE PROTOTYPES___________________*/



/*__________________PROGRAMMA___________________*/



/*---------------------------------------------------------------------------*
 *	setup_dsp:
 *	  opens device (dspdev; usually /dev/dsp)
 *	  sets fragment size (2 or 3 fragments of size 'fragsize').
 *	  sets speed ('speed')
 *	  sets mono/stereo ('channels')
 *	  sets sample size; 8 bit mulaw/8 bit unsigned/16 bit ('type')
 *---------------------------------------------------------------------------*/
int setup_dsp (char *dspdev, int speed, int log2_fragsize, int channels, int type) {
	int  value;
	int  format;
#ifdef USE_OSS
	int  snd_formats[] = { AFMT_MU_LAW, AFMT_U8, AFMT_S16_LE };
#endif
	char *snd_format_strings[] = { "mulaw", "unsigned 8 bit", "signed 16 bit" };

	DEB1(fprt(stderr, "setup dsp\n"));

	/* A few sanity checks, put this later in the code ? */
	speed = min(speed, 48000);
	speed = max(speed, 4000);
	type  = min(type, GSI_16BIT);
	type  = max(type, GSI_8BIT_MULAW);
	log2_fragsize = min(log2_fragsize, 13); /* max 8K buffer at the moment */

	// Fragment size
	DEB1(fprt(stderr, "%s: setting fragment to 2^%d\n", server_name, log2_fragsize));
	if (log2_fragsize==0) {
		/* Choose the best fragsize
		 *                   1  2  3  4  5  6  7  8  9 10 11 12 */
		int  base_log[12]= { 0, 1, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4 };
		log2_fragsize=base_log[(speed+2000)/4000 -1]+7+(channels>>1)+((type+1)>>2);
		/* e.g. 8000Hz mono 8 bit=256, 48000Hz, stereo 16 bit=8192 */
		DEB1(fprintf(stderr, "auto log2_fragsize=%d\n", log2_fragsize));
	}

#ifdef USE_OSS
//	value=0x00020000 | log2_fragsize; /* 2 fragments of size fragsize */ 
	value=0x00030000 | (log2_fragsize-1); /* 3 fragments of size fragsize/2 wroks better? */ 
	if (dsp>0 && log2_fragsize!=dsp_log2_frag) { close(dsp); dsp=-1; } // NB. This is needed for OSS as the fragsize cannot be adjusted after a read/write to the dsp has been done.
#endif /* USE_OSS */

	if (dsp<0) { // not yet opened
		dsp = open(dspdev, O_WRONLY);
		if (dsp < 1) {
			if (errno==EBUSY) {
				status_send_byte(GSI_CALLBACK_DSP_BUSY);
				perror("OPEN DSP"); //remove in final version
			}
			else
				fprintf (stderr, "%s: Couldn't open DSP %s\n", server_name, dspdev);
			return -1;
		}
	}

#ifdef USE_OSS
	ioctl(dsp, SNDCTL_DSP_SETFRAGMENT, &value);
#endif
#ifdef USE_DEV_AUDIO
#  warning /dev/audio: can we set fragment size?
#endif

	dsp_speed=speed;
	dsp_log2_frag=log2_fragsize;
	dsp_channels=channels;
	dsp_type=type;


	// Channels
	//if (ioctl(dsp, SNDCTL_DSP_CHANNELS, &dsp_channels)==-1) {
	value = dsp_channels-1; //0=mono, 1=stereo
#ifdef USE_OSS
	if (ioctl(dsp, SNDCTL_DSP_STEREO, &value)==-1) {
		perror("SNDCTL_DSP_STEREO"); //remove in final version
	}
#endif
#ifdef USE_DEV_AUDIO
#  warning /dev/audio: can we set stereo/mono?
	value=0; /* mono */
#endif
	if (dsp_channels==2 && value!=1) {
		fprintf(stderr, "DSP device doesn't support stereo, using mono.\n");
	}
	dsp_channels=value+1;
	DEB1(fprt(stderr, "channels=%d\n", value+1));

	{
	int first_dsp_type=dsp_type;
	for(;;) {
		// Sample size: 8 bit mulaw, 8 bit unsigned or 16 bit
#ifdef USE_OSS
		format = snd_formats[dsp_type-1];
		if (ioctl(dsp, SNDCTL_DSP_SETFMT, &format)!=-1) {
			if (format ==snd_formats[dsp_type-1])
				break;
		}
/*		perror("SNDCTL_DSP_SETFMT");*/
#endif
#ifdef USE_DEV_AUDIO
		if (dsp_type == GSI_8BIT_MULAW)
			break;
#endif
		fprintf(stderr, "DSP: audio format %s not supported\n", snd_format_strings[dsp_type-1]);
		dsp_type++;
		if (dsp_type>GSI_16BIT) dsp_type=GSI_8BIT_MULAW;
		if (dsp_type==first_dsp_type) {
			fprintf(stderr, "The DSP device doesn't support any of the formats\n");
			exit(1);
		}
	}
	}
	DEB1(fprt(stderr, "DSP format set to %s\n", snd_format_strings[dsp_type-1]));

	// Speed
	value=speed;
#ifdef USE_OSS
	if (ioctl(dsp, SNDCTL_DSP_SPEED, &value)) {
		fprintf (stderr, "%s: Couldn't set DSP rate\n", server_name);
	}
#endif
#ifdef USE_DEV_AUDIO
#  warning /dev/audio: can we set the rate ? (now using 8000Hz default)
	value=8000;
#endif
	DEB1(fprt(stderr, "DSP speed set at %d\n", value));
	dsp_speed=value;

	// Fragment size check
	dsp_fragsize = 0;
#ifdef USE_OSS
	ioctl(dsp, SNDCTL_DSP_GETBLKSIZE, &dsp_fragsize);
	if (!dsp_fragsize) {
		fprintf (stderr, "%s: Couldn't set fragment size.\n", server_name);
		exit(0); /* perhaps try another value? */
	}
#endif
#ifdef USE_DEV_AUDIO
#  warning /dev/audio: can we set fragsize ? (using 2<<log2_fragsize)
	dsp_fragsize=2<<log2_fragsize;
#endif
	DEB1(fprt(stderr, "server fragsize=%d\n", dsp_fragsize));

	dsp_bps= (dsp_type & 1)+1;

	dsp_samples= dsp_fragsize/(dsp_bps*dsp_channels);

// Dynamic allocation of the sound buffers, necessary?
// ---------------------------------------------------
//if (final!=NULL) free(final); if (premix!=NULL) free(premix); 
//   final= (byte *) malloc(dsp_fragsize);
//     /* only need to change 'final' is fragsize changes */
//   premix= (int *) malloc(dsp_samples*sizeof(int));
//     /* must change 'premix' if bytes or channels changes */


	DEB1(fprt(stderr, "server bps=%d, samples=%d, channels=%d, speed=%d\n", dsp_bps, dsp_samples, dsp_channels, dsp_speed));

	return dsp;
}

#if 0
void auto_dsp_update() {
	if (autorelease_dsp) {
		if (playnum>0 || channel_playnum>0) { /* we want to use the DSP */
			if (dsp<0)
				dsp=setup_dsp(dspdev, new_dsp_speed, new_dsp_log2_frag, new_dsp_channels, new_dsp_type);
		}
		else { /*we don't need the DSP any more */
			if (dsp>0) { close(dsp); dsp=-1; }
		}
	}
}
#endif


/*---------------------------------------------------------------------------*
 *	EOF dsp.c
 *---------------------------------------------------------------------------*/
